@@ -5,7 +5,7 @@
   $Author: matz $
   created at: Fri Aug  2 09:24:12 JST 1996
 
-  Copyright (C) 1995 Yukihiro Matsumoto
+  Copyright (C) 1995-1998 Yukihiro Matsumoto
 
 ************************************************/
 /* This module provides an interface to the RSA Data Security,
@@ -26,7 +26,7 @@ md5_update(obj, str)
     MD5_CTX *md5;
 
     Check_Type(str, T_STRING);
-    Get_Data_Struct(obj, MD5_CTX, md5);
+    Data_Get_Struct(obj, MD5_CTX, md5);
     MD5Update(md5, str->ptr, str->len);
 
     return Qnil;
@@ -38,7 +38,7 @@ md5_digest(obj)
     MD5_CTX *md5, ctx;
     unsigned char digest[16];
 
-    Get_Data_Struct(obj, MD5_CTX, md5);
+    Data_Get_Struct(obj, MD5_CTX, md5);
     ctx = *md5;
     MD5Final(digest, &ctx);
 
@@ -52,8 +52,8 @@ md5_clone(obj)
     VALUE clone;
     MD5_CTX *md5, *md5_new;
 
-    Get_Data_Struct(obj, MD5_CTX, md5);
-    obj = Make_Data_Struct(CLASS_OF(obj), MD5_CTX, 0, 0, md5_new);
+    Data_Get_Struct(obj, MD5_CTX, md5);
+    obj = Data_Make_Struct(CLASS_OF(obj), MD5_CTX, 0, free, md5_new);
     *md5_new = *md5;
 
     return obj;
@@ -61,6 +61,9 @@ md5_clone(obj)
 
 static VALUE
 md5_new(argc, argv, class)
+    int argc;
+    VALUE* argv;
+    VALUE class;
 {
     int i;
     VALUE arg, obj;
@@ -69,11 +72,12 @@ md5_new(argc, argv, class)
     rb_scan_args(argc, argv, "01", &arg);
     if (!NIL_P(arg)) Check_Type(arg, T_STRING);
 
-    obj = Make_Data_Struct(class, MD5_CTX, 0, 0, md5);
+    obj = Data_Make_Struct(class, MD5_CTX, 0, free, md5);
     MD5Init(md5);
     if (!NIL_P(arg)) {
 	md5_update(obj, arg);
     }
+    obj_call_init(obj);
 
     return obj;
 }
