@@ -3,7 +3,7 @@
   time.c -
 
   $Author: matz $
-  $Date: 1999/01/18 09:08:51 $
+  $Date: 1999/04/08 10:14:35 $
   created at: Tue Dec 28 14:31:59 JST 1993
 
   Copyright (C) 1993-1998 Yukihiro Matsumoto
@@ -396,6 +396,21 @@ time_hash(time)
     return INT2FIX(hash);
 }
 
+static VALUE
+time_clone(time)
+    VALUE time;
+{
+    VALUE obj;
+    struct time_object *tobj, *newtobj;
+
+    GetTimeval(time, tobj);
+    obj = Data_Make_Struct(0, struct time_object, 0, free, newtobj);
+    CLONESETUP(obj, time);
+    MEMCPY(newtobj, tobj, struct time_object, 1);
+
+    return obj;
+}
+
 static VALUE
 time_localtime(time)
     VALUE time;
